import subto.job as sjob
import copy
import pathlib


class OrcaJob():
    '''
    Class to generate submission scripts for ORCA calculation\n\n

    Submission script format is \n

    ...............\n
    SCHEDULER BLOCK\n
    ...............\n
    LOAD_BLOCK\n
    ...............\n
    PRE-ORCA_BLOCK\n
    ...............\n
    $(which orca) input_file > output_file\n
    ...............\n
    POST-ORCA_BLOCK\n
    ...............\n

    Attributes
    ----------
    input_file: pathlib.Path
        Orca input file for which a submission script is created
    output_file: pathlib.Path
        Orca output file name. Default is same as input with extension \n
        replaced by .out
    job_file: pathlib.Path
        Submission script file path. Default is same as input with extension \n
        replaced by scheduler job file extension e.g. '.slm'
    job_name: str
        Job name. Default is same as input without extension
    scheduler_name: str {'slurm'}
        Scheduler system name
    pre_orca: str
        Commands for pre-orca block
    post_orca: str
        Commands for post-orca block
    '''

    SUPPORTED_SCHEDULERS = [
        sjob.SlurmJob
    ]

    def __init__(self, input_file: str | pathlib.Path,
                 Job: sjob.Job = sjob.SlurmJob, **kwargs) -> None:

        self.input_file = input_file
        self.Job = Job

        # Set defaults
        self.output_file = self.input_file.with_suffix('.out')
        self.job_file = self.input_file.with_suffix(self.Job.EXTENSION)

        self.job_name = copy.copy(self.input_file.stem)
        self.load = ''
        self.pre_orca = ''
        self.post_orca = ''

        # If provided set attrs using kwargs
        for key, value in kwargs.items():
            setattr(self, key, value)
        return

    @property
    def input_file(self) -> pathlib.Path:
        return self._input_file

    @input_file.setter
    def input_file(self, value: str | pathlib.Path):
        if isinstance(value, str):
            self._input_file = pathlib.Path(value).resolve()
        elif isinstance(value, pathlib.Path):
            self._input_file = value.resolve()
        else:
            raise TypeError('input_file should be of type str or Path')
        return

    @property
    def input_file_relpath(self) -> pathlib.Path:
        return self.input_file.relative_to(pathlib.Path.cwd())

    @property
    def output_file(self) -> pathlib.Path:
        return self._output_file

    @output_file.setter
    def output_file(self, value: str | pathlib.Path):
        if isinstance(value, str):
            self._output_file = pathlib.Path(value).resolve()
        elif isinstance(value, pathlib.Path):
            self._output_file = value.resolve()
        else:
            raise TypeError('output_file should be of type str or Path')
        return

    @property
    def job_file(self) -> pathlib.Path:
        '''
        The submission script file path
        '''
        return self._job_file

    @job_file.setter
    def job_file(self, value: str | pathlib.Path):
        if isinstance(value, str):
            self._job_file = pathlib.Path(value).resolve()
        elif isinstance(value, pathlib.Path):
            self._job_file = value.resolve()
        else:
            raise TypeError('job_file should be of type str or Path')
        return

    @property
    def output_file_relpath(self) -> pathlib.Path:
        return self.output_file.relative_to(pathlib.Path.cwd())

    @property
    def Job(self) -> sjob.Job:
        return self._Job

    @Job.setter
    def Job(self, value: sjob.Job):
        if value not in self.SUPPORTED_SCHEDULERS:
            raise ValueError('Unsupported Job Class provided to OrcaJob')
        else:
            self._Job = value
        return

    @property
    def pre_orca(self):
        rvalue = copy.copy(self._pre_orca)
        rvalue = rvalue.replace('<input>', self.input_file.name)
        rvalue = rvalue.replace('<output>', self.output_file.name)
        rvalue = rvalue.replace('<stem>', self.input_file.stem)
        return rvalue

    @pre_orca.setter
    def pre_orca(self, value: str):
        self._pre_orca = value

    @property
    def post_orca(self):
        rvalue = copy.copy(self._post_orca)
        rvalue = rvalue.replace('<input>', self.input_file.name)
        rvalue = rvalue.replace('<output>', self.output_file.name)
        rvalue = rvalue.replace('<stem>', self.input_file.stem)
        return rvalue

    @post_orca.setter
    def post_orca(self, value: str):
        self._post_orca = value

    @property
    def orca_command(self):
        return f'$(which orca) {self.input_file.name} >> {self.output_file.name}' # noqa

    def write_script(self, verbose: bool = True, **kwargs) -> None:
        '''
        Writes submission script to file

        Parameters
        ----------
        verbose: bool, default True
            If True, jobscript location is written to screen
        **kwargs
            Keyword arguments passed to `subto.job.Job` constructor
        '''

        if len(self.load):
            self.load += '\n\n'
        if len(self.pre_orca):
            self.pre_orca += '\n\n'
        if len(self.post_orca):
            self.post_orca = '\n' + self.post_orca + '\n\n'

        self.pre_orca += f'\ndate > {self.output_file.name}\n'
        self.pre_orca += f'hostname >> {self.output_file.name}\n\n'

        # Generate content of script
        content = '{}{}{}\n{}'.format(
            self.load,
            self.pre_orca,
            self.orca_command,
            self.post_orca
        )

        # And use scheduler object to submission system job object
        job = self.Job(
            file=self.job_file,
            job_name=self.job_name,
            content_block=content,
            **kwargs
        )

        # Write script
        job.write_script(verbose=verbose)

        return



import numpy as np
from numpy.typing import NDArray, ArrayLike
import pandas as pd

from .exceptions import DataFormattingError
from . import constants as const
from . import extractor as oe
from . import utils as ut


class InfraredSpectrum():
    '''
    Stores infrared spectrum data

    Parameters
    ----------
    x_grid: NDArray
        Grid of x values (wavenumbers)
    y_unit: str
        Unit of y values, e.g. 'cm^-1 mol^-1 L'
    y_label: str
        Label for y axis
    y_label_mathmode: str
        Label for y axis using LaTeX mathmode formatting
    spectrum: NDArray
        Infrared spectrum values at each point in x_grid
    fwhm: float
        Full width at half maximum used in spectrum generation
    comment: str, default=''
        Comment or metadata for the spectrum
    x_reversed: bool, default False
        Whether the x axis is reversed\n
        By default, False - increasing x values from left to right

    Attributes
    ----------
    x_grid: NDArray
        Grid of x values (wavenumbers)
    y_unit: str
        Unit of y values, e.g. 'cm^-1 mol^-1 L'
    y_label: str
        Label for y axis
    y_label_mathmode: str
        Label for y axis using LaTeX mathmode formatting
    y_values: NDArray
        Infrared spectrum values at each point in x_grid
    fwhm: float
        Full width at half maximum used in spectrum generation
    comment: str
        Comment or metadata for the spectrum
    '''

    def __init__(self, x_grid: NDArray, y_unit: str, y_label: str,
                 y_label_mathmode: str, y_values: NDArray, fwhm: float,
                 comment: str = '', x_reversed: bool = False) -> None:
        self.x_grid = x_grid
        self.y_unit = y_unit
        self.y_label = y_label
        self.y_label_mathmode = y_label_mathmode
        self.fwhm = fwhm
        self.y_values = y_values
        self.comment = comment
        self.x_reversed = x_reversed
        return


class FrequencyData():
    def __init__(self, wavenumber: ArrayLike, linear_absorbance: ArrayLike):

        self.wavenumber = wavenumber
        self.linear_absorbance = linear_absorbance
        self.spectrum: None | InfraredSpectrum = None
        return

    @property
    def spectrum(self) -> InfraredSpectrum:
        '''Infrared spectrum object'''
        if self._spectrum is None:
            raise ValueError(
                'Spectrum has not been generated yet\n'
                'Call generate_spectrum() first.'
            )
        return self._spectrum

    @spectrum.setter
    def spectrum(self, spectrum: InfraredSpectrum) -> None:
        if not isinstance(spectrum, InfraredSpectrum | None):
            raise TypeError('spectrum must be an InfraredSpectrum object')
        self._spectrum = spectrum
        return

    @classmethod
    def from_extractor(cls, extractor: oe.FrequencyExtractor) -> list['FrequencyData']: # noqa
        '''
        Creates FrequencyData object(s) from data extractor

        Parameters
        ----------
        extractor: oe.FrequencyExtractor
            Data extractor object containing transition data

        Returns
        -------
        list['FrequencyData']
            FrequencyData objects, one for each section in extractor
        '''

        all_freq_data = [
            cls.from_extractor_dataset(
                dataset
            )
            for dataset in extractor.data
        ]

        return all_freq_data

    @classmethod
    def from_extractor_dataset(cls,
                               dataset: dict[str, list[int | float]]) -> 'FrequencyData': # noqa
        '''
        Creates FrequencyData object(s) from extractor data

        Parameters
        ----------
        dataset: dict[str, list[int | float]]
            Dataset from orto FrequencyExtractor.data

        Returns
        -------
        FrequencyData
            FrequencyData object for dataset
        '''

        wavenumber = dataset['energy (cm^-1)']
        # Integrated Linear Absorbance = Epsilon in Orca output file
        # Since it is computed with a dirac delta lineshape
        linear_absorbance = dataset['epsilon (L mol^-1 cm^-1)']

        return cls(wavenumber, linear_absorbance)

    def generate_spectrum(self, fwhm: float, lineshape: str, x_min: float,
                          x_max: float, num_points: int, comment: str = '',
                          normalise: bool = False, y_type: str = 'absorbance',
                          x_reversed: bool = False) -> InfraredSpectrum:
        '''
        Generates infrared (absorption or transmittance) spectrum
        using specified broadening.

        Parameters
        ----------
        fwhm: float
            Full width at half maximum for broadening\n
            in same units as x_type. Applied to each transition.
        lineshape: str
            Lineshape function to use: 'gaussian' or 'lorentzian'
        x_min: float | str
            Minimum x value for spectrum (eV, nm, or cm^-1)
        x_max: float | str
            Maximum x value for spectrum (eV, nm, or cm^-1)
        num_points: int
            Number of points in the spectrum grid
        comment: str
            Comment to add to spectrum metadata
        normalise: bool, default False
            If True, normalises y data by largest value.\n
            If y_type='absorbance' -> normalised absorbance 0->1\n
            If y_type='transmittance' -> % transmittance 0->100%
        y_type: str, default 'absorbance'
            y data type, either 'absorbance' or 'transmittance'\n
            n.b. that this is linear (decadic) absorbance, not napierian\n
            n.b. that y_type='transmittance' also enables normalisation.
        x_reversed: bool
            Whether to reverse the x axis (e.g. for wavelength spectra)\n
            By default, False - increasing x values from left to right

        Returns
        -------
        None
        '''

        if len(self.wavenumber) == 0 or len(self.linear_absorbance) == 0:
            raise DataFormattingError(
                'No data available to generate spectrum.'
            )
        elif len(self.wavenumber) != len(self.linear_absorbance):
            raise DataFormattingError(
                'Wavenumbers and Linear Absorbances must have the same length.'
            )

        # Create grid for spectrum
        x_grid = np.linspace(x_min, x_max, num_points)
        spectrum = np.zeros_like(x_grid)

        ls_func = {
            'gaussian': ut.gaussian,
            'lorentzian': ut.lorentzian
        }

        # Spectrum as sum of signals. Always computed in wavenumbers.
        y_vals = np.sum([
            ls_func[lineshape](x_grid, fwhm, wavenumber, a)
            for wavenumber, a in zip(self.wavenumber, self.linear_absorbance)
        ], axis=0)

        # Normalise Absorbance
        if y_type == 'absorbance':
            if normalise:
                y_vals /= np.max(y_vals)
                y_unit = ''
                y_label = 'Normalised absorbance'
                y_label_mathmode = 'Normalised absorbance'
            else:
                y_unit = '`cm mol^-1 eV^-1'
                y_label = r'ε (cm⁻¹ mol⁻¹ L))'
                y_label_mathmode = r'$\epsilon$ (cm$^\mathregular{-1}$ mol$^\mathregular{-1}$ L)'  # noqa
        elif y_type == 'transmittance':
            y_vals /= np.max(y_vals)
            y_vals = 100 * 10 ** (- y_vals)
            y_unit = ''
            y_label = 'Transmittance / %'
            y_label_mathmode = 'Transmittance / %'
        else:
            raise ValueError(
                f'Unknown y_type {y_type}, should be absorbance or transmittance' # noqa
            )

        spectrum = InfraredSpectrum(
            x_grid=x_grid,
            y_unit=y_unit,
            y_label=y_label,
            y_label_mathmode=y_label_mathmode,
            y_values=y_vals,
            fwhm=fwhm,
            comment=comment,
            x_reversed=x_reversed
        )

        self.spectrum = spectrum

        return


class AbsorptionSpectrum():
    '''
    Stores absorption spectrum data

    Parameters
    ----------
    x_grid: NDArray
        Grid of x values (energies, wavelengths, or wavenumbers)
    x_unit: str
        Unit of x values: 'eV', 'nm', or 'cm^-1'
    x_label: str
        Label for x axis
    x_label_mathmode: str
        Label for x axis using LaTeX mathmode formatting
    y_unit: str
        Unit of y values, e.g. 'cm^-1 mol^-1 L'
    y_label: str
        Label for y axis
    y_label_mathmode: str
        Label for y axis using LaTeX mathmode formatting
    spectrum: NDArray
        Absorption spectrum values at each point in x_grid
    fwhm: float
        Full width at half maximum used in spectrum generation
    comment: str, default=''
        Comment or metadata for the spectrum
    x_reversed: bool, default False
        Whether the x axis is reversed (e.g. for wavelength spectra)\n
        By default, False - increasing x values from left to right

    Attributes
    ----------
    x_grid: NDArray
        Grid of x values (energies, wavelengths, or wavenumbers)
    x_unit: str
        Unit of x values: 'eV', 'nm', or 'cm^-1'
    x_label: str
        Label for x axis
    x_label_mathmode: str
        Label for x axis using LaTeX mathmode formatting
    y_unit: str
        Unit of y values, e.g. 'cm^-1 mol^-1 L'
    y_label: str
        Label for y axis
    y_label_mathmode: str
        Label for y axis using LaTeX mathmode formatting
    y_values: NDArray
        Absorption spectrum values at each point in x_grid
    fwhm: float
        Full width at half maximum used in spectrum generation
    comment: str
        Comment or metadata for the spectrum
    '''

    def __init__(self, x_grid: NDArray, x_unit: str, x_label: str,
                 x_label_mathmode: str, y_unit: str, y_label: str,
                 y_label_mathmode: str, y_values: NDArray, fwhm: float,
                 comment: str = '', x_reversed: bool = False) -> None:
        self.x_grid = x_grid
        self.x_unit = x_unit
        self.x_label = x_label
        self.x_label_mathmode = x_label_mathmode
        self.y_unit = y_unit
        self.y_label = y_label
        self.y_label_mathmode = y_label_mathmode
        self.fwhm = fwhm
        self.y_values = y_values
        self.comment = comment
        self.x_reversed = x_reversed
        return


class AbsorptionData():
    '''
    Stores absorption data for a given set of transitions

    Parameters
    ----------
    energies: array_like
        Transition energies in eV
    osc_strengths: array_like
        Oscillator strengths for each transition
    operator: str
        Type of operator used to calculate transitions

    Attributes
    ----------
    energies: list
        Transition energies in eV
    osc_strengths: list
        Oscillator strengths for each transition
    operator: str
        Type of operator used to calculate transitions
    spectrum: AbsorptionSpectrum
        Spectrum object containing spectrum curve
    wavelengths: list
        Transition wavelengths in nm
    wavenumbers: list
        Transition wavenumbers in cm^-1
    '''

    def __init__(self, energies: ArrayLike, osc_strengths: ArrayLike,
                 operator: str) -> None:
        self.energies = energies
        self.osc_strengths = osc_strengths
        self.operator = operator
        self.spectrum: None | AbsorptionSpectrum = None
        return

    @property
    def osc_strengths(self) -> list:
        '''Oscillator strengths for each transition'''
        return self._osc_strengths

    @osc_strengths.setter
    def osc_strengths(self, osc_strengths: ArrayLike) -> None:
        self._osc_strengths = list(osc_strengths)
        return

    @property
    def energies(self) -> list:
        '''Transition energies in eV'''
        return self._energies

    @energies.setter
    def energies(self, energies: ArrayLike) -> None:
        self._energies = list(energies)
        return

    @property
    def wavelengths(self) -> list:
        '''Transition wavelengths in nm'''
        return [const.EV_TO_NM / energy for energy in self.energies]

    @property
    def wavenumbers(self) -> list:
        '''Transition wavenumbers in cm^-1'''
        return [energy * const.EV_TO_WAVENUMBER for energy in self.energies]

    @property
    def spectrum(self) -> AbsorptionSpectrum:
        '''Absorption spectrum object'''
        if self._spectrum is None:
            raise ValueError(
                'Spectrum has not been generated yet\n'
                'Call generate_spectrum() first.'
            )
        return self._spectrum

    @spectrum.setter
    def spectrum(self, spectrum: AbsorptionSpectrum) -> None:
        if not isinstance(spectrum, AbsorptionSpectrum | None):
            raise TypeError('spectrum must be an AbsorptionSpectrum object')
        self._spectrum = spectrum
        return

    def trim_to_n(self, n_transitions: int) -> None:
        '''
        Trims data to given number of transitions

        Parameters
        ----------
        n_transitions: int
            Number of transitions to keep from the start of the data

        Returns
        -------
        None
        '''

        self.energies = self.energies[:n_transitions]
        self.osc_strengths = self.osc_strengths[:n_transitions]

        self.spectrum = None

        return

    @classmethod
    def from_extractor(cls, extractor: oe.AbsorptionExtractor,
                       remove_zero_osc: float = 1E-4) -> list['AbsorptionData']: # noqa
        '''
        Creates AbsorptionData object(s) from data extractor

        Parameters
        ----------
        extractor: oe.AbsorptionExtractor
            Data extractor object containing transition data
        remove_zero_osc: float, default=1E-4
            Remove transitions with oscillator strength below this value

        Returns
        -------
        list['AbsorptionData']
            AbsorptionData objects, one for each section in extractor
        '''

        all_abs_data = [
            cls.from_extractor_dataset(
                dataset,
                extractor.operator_type,
                remove_zero_osc=remove_zero_osc
            )
            for dataset in extractor.data
        ]

        return all_abs_data

    @classmethod
    def from_extractor_dataset(cls,
                               dataset: dict[str, list[int | float]],
                               operator: str,
                               remove_zero_osc: float = 1E-7) -> 'AbsorptionData': # noqa
        '''
        Creates AbsorptionData object(s) from data extractor

        Parameters
        ----------
        dataset: dict[str, list[int | float]]
            Dataset from orto AbsorptionExtractor.data
        operator: str
            Type of operator used to calculate transitions
        remove_zero_osc: float, default=1E-7
            Remove transitions with oscillator strength below this value

        Returns
        -------
        AbsorptionData
            AbsorptionData object for dataset
        '''

        energies = dataset['energy (ev)']
        osc_strengths = dataset['fosc']

        # Remove transitions with zero oscillator strength from
        # beginning and end of spectrum
        osc_strengths = np.array(osc_strengths)

        # Find first non-zero oscillator strength
        try:
            first_nonzero = np.where(osc_strengths > remove_zero_osc)[0][0]
        except IndexError:
            raise DataFormattingError(
                'No transitions with oscillator strength above '
                f'{remove_zero_osc} found.'
            )
        # Find last non-zero oscillator strength
        last_nonzero = np.where(osc_strengths > remove_zero_osc)[0][-1]

        # Trim data
        energies = energies[first_nonzero:last_nonzero + 1]
        osc_strengths = osc_strengths[first_nonzero:last_nonzero + 1]

        return cls(energies, osc_strengths, operator)

    def generate_spectrum(self, fwhm: float, lineshape: str, x_min: float,
                          x_max: float, num_points: int,
                          x_type: str = 'energy',
                          comment: str = '',
                          x_reversed: bool = False) -> AbsorptionSpectrum:
        '''
        Generates absorption spectrum using Gaussian broadening

        Parameters
        ----------
        fwhm: float
            Full width at half maximum for Gaussian broadening\n
            in same units as x_type. Applied to each transition.
        lineshape: str
            Lineshape function to use: 'gaussian' or 'lorentzian'
        x_min: float | str
            Minimum x value for spectrum (eV, nm, or cm^-1)
        x_max: float | str
            Maximum x value for spectrum (eV, nm, or cm^-1)
        num_points: int
            Number of points in the spectrum grid
        x_type: str
            Type of x values: \n
              'energy' (eV)\n
              'wavelength' (nm)\n
              'wavenumber' (cm^-1)
        comment: str
            Comment to add to spectrum metadata
        x_reversed: bool
            Whether to reverse the x axis (e.g. for wavelength spectra)\n
            By default, False - increasing x values from left to right

        Returns
        -------
        None
        '''

        if len(self.energies) == 0 or len(self.osc_strengths) == 0:
            raise DataFormattingError(
                'No transition data available to generate spectrum.'
            )
        elif len(self.energies) != len(self.osc_strengths):
            raise DataFormattingError(
                'Energies and oscillator strengths must have the same length.'
            )

        # Set labels and x values based on x_type
        if x_type == 'energy':
            x_unit = 'eV'
            x_label = 'Energy (eV)'
            x_label_mathmode = 'Energy (eV)'
            x_vals = self.energies
        elif x_type == 'wavelength':
            x_unit = 'nm'
            x_label = 'Wavelength (nm)'
            x_label_mathmode = 'Wavelength (nm)'
            x_vals = self.wavelengths
        elif x_type == 'wavenumber':
            x_unit = 'cm^-1'
            x_label = r'Wavenumber (cm⁻¹)'
            x_label_mathmode = r'Wavenumber (cm$\mathregular{^{-1}}$)'
            x_vals = self.wavenumbers
        else:
            raise DataFormattingError(f'Invalid x_type: {x_type}')

        y_unit = 'cm mol^-1 eV^-1'
        y_label = r'ε (cm⁻¹ mol⁻¹ L))'
        y_label_mathmode = r'$\epsilon$ (cm$^\mathregular{-1}$ mol$^\mathregular{-1}$ L)'  # noqa

        # Create grid for spectrum
        x_grid = np.linspace(x_min, x_max, num_points)
        spectrum = np.zeros_like(x_grid)

        # Conversion from oscillator strength to napierian integrated
        # absorption coefficient
        # This is the value of A for a harmonically oscillating electron
        A_elec = 2.31E8
        #
        A_logs = [fosc * A_elec for fosc in self.osc_strengths]

        ls_func = {
            'gaussian': ut.gaussian,
            'lorentzian': ut.lorentzian
        }

        # Spectrum as sum of signals in given unit
        y_vals = np.sum([
            ls_func[lineshape](x_grid, fwhm, x_value, A_log)
            for x_value, A_log in zip(x_vals, A_logs)
        ], axis=0)

        # Transform y values into epsilon in cm^-1 mol^-1 L
        if x_type == 'energy':
            y_vals /= const.EV_TO_WAVENUMBER
        elif x_type == 'wavelength':
            # calc x grid in wavenumbers
            xgwn = 1E7 / x_grid
            # Calculate jacobian for non linear transformation
            jacobian = (1E7 / xgwn**2)  # d(nm)/d(cm^-1)
            # obtain y values in wavenumbers
            y_vals = y_vals * jacobian

        spectrum = AbsorptionSpectrum(
            x_grid=x_grid,
            x_unit=x_unit,
            x_label=x_label,
            x_label_mathmode=x_label_mathmode,
            y_unit=y_unit,
            y_label=y_label,
            y_label_mathmode=y_label_mathmode,
            y_values=y_vals,
            fwhm=fwhm,
            comment=comment,
            x_reversed=x_reversed
        )

        self.spectrum = spectrum

        return

    def save_spectrum_data(self, filename: str, comments: str = '') -> None:
        '''
        Saves absorption spectrum data to CSV file

        Parameters
        ----------
        filename: str
            Name of output CSV file
        comments: str
            Comments to add at the top of the file

        Returns
        -------
        None
        '''

        if self.spectrum is None:
            raise ValueError(
                'Spectrum has not been generated yet\n'
                'Call generate_spectrum() first.'
            )

        # Select x values based on x_type
        if self.spectrum.x_label.split()[0].lower() == 'energy':
            x_vals = self.spectrum.x_grid
        elif self.spectrum.x_label.split()[0].lower() == 'wavelength':
            x_vals = const.EV_TO_NM / self.spectrum.x_grid
        elif self.spectrum.x_label.split()[0].lower() == 'wavenumber':
            x_vals = self.spectrum.x_grid * const.EV_TO_WAVENUMBER

        if len(comments):
            header = comments + f'\n{self.spectrum.x_label}, ε (cm⁻¹ mol⁻¹ L))'
        else:
            header = f'{self.spectrum.x_label}, ε (cm⁻¹ mol⁻¹ L))'

        # Save to CSV
        np.savetxt(
            filename,
            np.column_stack((x_vals, self.spectrum.y_values)),
            delimiter=',',
            header=header,
            comments='#',
        )

    def save_transition_data(self, filename: str, comments: str = '') -> None:
        '''
        Saves transition data to CSV file

        Parameters
        ----------
        filename: str
            Name of output CSV file
        comments: str
            Comments to add at the top of the file

        Returns
        -------
        None
        '''

        if len(comments):
            header = comments + '\nEnergy (eV),Wavelength (nm),Wavenumber (cm⁻¹),Oscillator Strength'  # noqa
        else:
            header = 'Energy (eV),Wavelength (nm),Wavenumber (cm⁻¹),Oscillator Strength'  # noqa

        np.savetxt(
            filename,
            np.column_stack((
                self.energies,
                self.wavelengths,
                self.wavenumbers,
                self.osc_strengths
            )),
            delimiter=',',
            header=header,
            comments='#',
        )


class SusceptibilityData():
    '''
    Stores magnetic susceptibility data

    Parameters
    ----------
    chis: array_like
        Susceptibility values in cm^3 K mol^-1
    temperatures: array_like
        Temperatures in K
    field: float
        Static magnetic field in Oe

    Attributes
    ----------
    chis: ndarray of floats
        Susceptibility values in cm^3 K mol^-1
    temperature: ndarray of floats
        Temperatures in K
    field: float
        Static magnetic field in Oe
    '''

    def __init__(self, chis: ArrayLike, temperatures: ArrayLike,
                 field: float) -> None:
        self.chis = chis
        self.temperatures = temperatures
        self.field = field
        return

    @property
    def chis(self) -> NDArray:
        '''Magnetic susceptibility values in cm^3 K mol^-1'''
        return self._chis

    @chis.setter
    def chis(self, chis: ArrayLike) -> None:
        self._chis = np.asarray(chis)
        return

    @property
    def temperatures(self) -> NDArray:
        '''Temperature values in K`'''
        return self._temperatures

    @temperatures.setter
    def temperatures(self, temperatures: ArrayLike) -> None:
        self._temperatures = np.asarray(temperatures)
        return

    @property
    def field(self) -> float:
        '''Static magnetic field in Oe'''
        return self._field

    @field.setter
    def field(self, field: float) -> None:
        self._field = float(field)
        return

    @classmethod
    def from_extractor(cls, extractor: oe.AbsorptionExtractor) -> list['SusceptibilityData']: # noqa
        '''
        Creates SusceptibilityData object(s) from data extractor

        Parameters
        ----------
        extractor: oe.SusceptibilityExtractor
            Data extractor object containing susceptibility data

        Returns
        -------
        list['SusceptibilityData']
            SusceptibilityData objects, one for each section in extractor
        '''

        all_abs_data = [
            cls.from_extractor_dataset(
                dataset
            )
            for dataset in extractor.data
        ]

        return all_abs_data

    @classmethod
    def from_extractor_data(cls, data: list[pd.DataFrame]) -> list['SusceptibilityData']: # noqa
        '''
        Creates SusceptibilityData object(s) extractor.data attribute

        Parameters
        ----------
        data: list[pd.DataFrame]
            List of dataframes from orto SusceptibilityExtractor.data attribute

        Returns
        -------
        list['SusceptibilityData']
            SusceptibilityData objects, one for each dataframe in data
        '''

        all_data = []

        # Generate a susceptibility data object for each dataframe
        # and static field value
        for dataset in data:
            # Get unique fields and split indices
            ufields, splindices = np.unique(
                dataset['Static Field (Gauss)'],
                return_index=True
            )
            splindices = list(splindices) + [len(dataset)]

            for it, field in enumerate(ufields):
                subdata = dataset[splindices[it]:splindices[it + 1]]

                temperatures = subdata['Temperature (K)']
                chis = subdata['chi*T (cm3*K/mol)'] / temperatures
                all_data.append(
                    cls(chis, temperatures, field)
                )

        return all_data



import matplotlib.pyplot as plt
from matplotlib.ticker import AutoMinorLocator
import numpy as np
from numpy.typing import ArrayLike
import itertools
import pathlib

from . import utils as ut
from . import data


def set_axlims(ax: plt.Axes, type: str, limits: list[float | str]) -> None:
    '''
    Sets axis limits of specified axis given a set of limit\n
    values or string 'auto'
    '''

    if type == 'y':
        gmethod = ax.get_ylim
        smethod = ax.set_ylim
    elif type == 'x':
        gmethod = ax.get_xlim
        smethod = ax.set_xlim
    else:
        raise ValueError(f'Unknown axis type {type}')

    _limits = limits.copy()
    if _limits[0] != _limits[1]:
        # Lower limit either auto or specified
        if isinstance(_limits[0], str):
            if _limits[0] == 'auto':
                _limits[0] = gmethod()[0]
            elif ut.is_floatable(_limits[0]):
                _limits[0] = float(_limits[0])
            else:
                raise ValueError(f'Invalid {type} limit value: {_limits[0]}')
        # Upper limit either auto or specified
        if isinstance(_limits[1], str):
            if _limits[1] == 'auto':
                _limits[1] = gmethod()[1]
            elif ut.is_floatable(_limits[1]):
                _limits[1] = float(_limits[1])
            else:
                raise ValueError(f'Invalid {type} limit value: {_limits[1]}')
        # Set values
        smethod([_limits[0], _limits[1]])

    return


def plot_absorption_spectrum(abs_data: data.AbsorptionData,
                             linecolor: str = 'black',
                             stickcolor: str = 'black',
                             xlim: list[float] = ['auto', 'auto'],
                             ylim: list[float] = [0., 'auto'],
                             x_shift: float = 0.,
                             normalise: bool = False,
                             osc_style: str = 'separate',
                             fig: plt.Figure = None, ax: plt.Axes = None,
                             oax: plt.Axes = None, save: bool = False,
                             save_name: str = 'absorption_spectrum.png',
                             show: bool = False,
                             verbose: bool = True,
                             window_title: str = 'Absorption Spectrum',
                             legend: bool = True) -> tuple[plt.Figure, plt.Axes, plt.Axes | None]: # noqa
    '''
    Plots an absorption spectrum from AbsorptionSpectrum object.\n

    Parameters
    ----------
    abs_data: data.AbsorptionData
        Data object containing absorption data and spectrum
    linecolor: str, default 'black'
        Color of the continuous spectrum line
    stickcolor: str, default 'black'
        Color of the oscillator strength sticks
    xlim: list[float], default ['auto', 'auto']
        Minimum and maximum x-values to plot
    ylim: list[float | str], default [0., 'auto']
        Minimum and maximum y-values to plot
    normalise: bool, default False
        If True, normalise the absorption spectrum to the maximum value.
    osc_style: str, default 'separate'
        Style of oscillator strength plots:
        - 'separate': plots oscillator strengths as stems on separate axis
        - 'combined': plots oscillator strengths on intensity axis
        - 'off': does not plot oscillator strengths
    fig: plt.Figure | None, optional
        If provided, uses this Figure object for plotting
    ax: plt.Axes | None, optional
        If provided, uses this Axis object for plotting spectrum
    oax: plt.Axes | None, optional
        If provided, uses this Axis object for plotting oscillator strengths
    save: bool, default False
        If True, plot is saved to save_name
    save_name: str | pathlib.Path, default 'absorption_spectrum.png'
        If save is True, plot is saved to this location/filename
    show: bool, default False
        If True, plot is shown on screen
    verbose: bool, default True
        If True, plot file location is written to terminal
    window_title: str, default 'Absorption Spectrum'
        Title of figure window, not of plot
    legend: bool, default True
        If True, a legend is added to the plot
    Returns
    -------
    plt.Figure
        Matplotlib Figure object
    list[plt.Axes]
        Matplotlib Axis object for main plot followed by\n
        Matplotlib Axis object for twinx oscillator strength axis
    '''
    save_name = pathlib.Path(save_name)

    if abs_data.spectrum is None:
        raise ValueError('AbsorptionData object does not contain spectrum')

    if fig is None or ax is None:
        width = 4
        width_cm = 10.4
        golden = (1 + np.sqrt(5))/2
        fig, ax = plt.subplots(
            1,
            1,
            num=window_title,
            figsize=(width, width / golden)
        )
    else:
        width = fig.get_size_inches()[0]
        width_cm = width * 2.54

    if oax is None and osc_style == 'separate':
        oax = ax.twinx()

    if not isinstance(xlim, list):
        raise ValueError('`xlim` must be a list of values')

    # Set x limits if specified
    set_axlims(ax, 'x', xlim)

    # Normalisation of spectrum
    if normalise:
        # Normalise the spectrum to the maximum value
        _y_values = abs_data.spectrum.y_values / np.max(abs_data.spectrum.y_values) # noqa
        _osc = abs_data.osc_strengths / np.max(abs_data.osc_strengths)
        ax.set_ylabel('Normalised Absorption')
    else:
        _y_values = abs_data.spectrum.y_values
        _osc = abs_data.osc_strengths
        ax.set_ylabel(abs_data.spectrum.y_label_mathmode)

    if abs_data.spectrum.x_label.split()[0].lower() == 'wavelength':
        stick_x_values = [
            wavelength + x_shift
            for wavelength in abs_data.wavelengths
        ]
    elif abs_data.spectrum.x_label.split()[0].lower() == 'energy':
        stick_x_values = [
            energy + x_shift
            for energy in abs_data.energies
        ]
    elif abs_data.spectrum.x_label.split()[0].lower() == r'wavenumber':
        stick_x_values = [
            wavenumber + x_shift
            for wavenumber in abs_data.wavenumbers
        ]

    if abs_data.spectrum.x_reversed:
        ax.invert_xaxis()

    x_values = abs_data.spectrum.x_grid + x_shift

    # Main spectrum
    ax.plot(
        x_values,
        _y_values,
        color=linecolor,
        lw=1.1,
        label=abs_data.spectrum.comment
    )

    if osc_style == 'separate':
        # Oscillator strength twin axis
        _, stemlines, _ = oax.stem(
            stick_x_values,
            _osc,
            basefmt=' ',
            markerfmt=' ',
            linefmt=stickcolor
        )
        plt.setp(stemlines, 'linewidth', 1.1)
        oax.yaxis.set_minor_locator(AutoMinorLocator())
        if normalise:
            oax.set_ylabel(r'Normalised $f_\mathregular{osc}$')
        else:
            oax.set_ylabel(r'$f_\mathregular{osc}$')
        oax.set_ylim([0., oax.get_ylim()[1]])
    elif osc_style == 'combined':
        ax.stem(
            stick_x_values,
            abs_data.osc_strengths/np.max(abs_data.osc_strengths) * np.max(_y_values), # noqa
            basefmt=' ',
            markerfmt=' ',
            linefmt=stickcolor
        )
        oax.set_yticks([])
        oax.spines['right'].set_visible(False)
        oax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
    else:
        # No oscillator strength plot
        plt.subplots_adjust(right=0.2)
        oax.set_yticks([])
        oax.spines['right'].set_visible(False)
        oax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)

    # Set y limits if specified
    set_axlims(ax, 'y', ylim)

    ax.xaxis.set_minor_locator(AutoMinorLocator())
    ax.yaxis.set_minor_locator(AutoMinorLocator())

    ax.set_xlabel(abs_data.spectrum.x_label_mathmode)

    fig.tight_layout()

    if legend:
        ax.legend()

    if save:
        plt.savefig(save_name, dpi=500)
        if verbose:
            ut.cprint(
                f'\nSpectrum saved to\n {save_name}',
                'cyan'
            )
            ut.cprint(
                f'Use width={width:.1f} in. or {width_cm:.1f} cm',
                'cyan'
            )

    if show:
        plt.show()

    return fig, ax, oax


def plot_susceptibility(susc_data: data.SuscseptibilityData,
                        y_style: str = 'XT',
                        label: str = None,
                        linecolor: str = 'black',
                        xlim: list[float] = ['auto', 'auto'],
                        ylim: list[float] = [0., 'auto'],
                        fig: plt.Figure = None, ax: plt.Axes = None,
                        save: bool = False,
                        save_name: str = 'magnetic_susceptibility.png',
                        show: bool = False,
                        verbose: bool = True,
                        window_title: str = 'Magnetic Susceptibility',
                        legend: bool = True) -> tuple[plt.Figure, plt.Axes]:
    '''
    Plots magnetic susceptibility data from SusceptibilityData object

    Parameters
    ----------
    susc_data: data.SusceptibilityData
        Data object containing susceptibility data
    y_style: str, default 'XT'
        Quanity to plot on y axis, one of\n
        'XT': chi*T\n
        'X': chi\n
        '1/X': 1/chi
    label: str
        Label for this trace used in legend
    linecolor: str, default 'black'
        Color of the data line
    xlim: list[float], default ['auto', 'auto']
        Minimum and maximum x-values to plot
    ylim: list[float | str], default [0., 'auto']
        Minimum and maximum y-values to plot
    fig: plt.Figure | None, optional
        If provided, uses this Figure object for plotting
    ax: plt.Axes | None, optional
        If provided, plots data to this axis
    save: bool, default False
        If True, plot is saved to save_name
    save_name: str | pathlib.Path, default 'magnetic_susceptibility.png'
        If save is True, plot is saved to this location/filename
    show: bool, default False
        If True, plot is shown on screen
    verbose: bool, default True
        If True, plot file location is written to terminal
    window_title: str, default 'Magnetic Susceptibility'
        Title of figure window, not of plot
    legend: bool, default True
        If True, a legend is added to the plot

    Returns
    -------
    plt.Figure
        Matplotlib Figure object
    plt.Axes
        Matplotlib Axis object for main plot
    '''

    save_name = pathlib.Path(save_name)

    if fig is None or ax is None:
        width = 4
        width_cm = 10.4
        golden = (1 + np.sqrt(5))/2
        fig, ax = plt.subplots(
            1,
            1,
            num=window_title,
            figsize=(width, width / golden)
        )
    else:
        width = fig.get_size_inches()[0]
        width_cm = width * 2.54

    if y_style == 'XT':
        y_data = susc_data.chis * susc_data.temperatures
    elif y_style == 'X':
        y_data = susc_data.chis
    elif y_style == '1/X':
        y_data = 1 / susc_data.chis
    else:
        raise ValueError(f'Unknown y_style {y_style}')

    if label is None:
        label = f'Calculated ($H$ = {susc_data.field:.1f} Oe)'

    ax.plot(
        susc_data.temperatures,
        y_data,
        label=label,
        lw=1.5,
        color=linecolor
    )

    # Set x and y limits
    set_axlims(ax, 'x', xlim)
    set_axlims(ax, 'y', ylim)

    ax.set_xlabel('$T$ (K)')
    ax.set_ylabel(r'$\chi T\,\mathregular{cm^3\,K\,mol^{-1}}$')

    ax.xaxis.set_minor_locator(AutoMinorLocator())
    ax.yaxis.set_minor_locator(AutoMinorLocator())

    fig.tight_layout()

    if save:
        plt.savefig(save_name, dpi=500)
        if verbose:
            ut.cprint(
                f'\nSpectrum saved to\n {save_name}',
                'cyan'
            )
            ut.cprint(
                f'Use width={width:.1f} in. or {width_cm:.1f} cm',
                'cyan'
            )

    if show:
        plt.show()

    return fig, ax


def plot_ir_spectrum(ir_spec: data.InfraredSpectrum, fig: plt.figure = None,
                     ax: plt.axes = None, color: str = 'k',
                     line_style: str = '-', save: bool = False,
                     save_name: str = 'ir_spectrum.png', show: bool = False,
                     verbose: bool = True,
                     window_title: str = 'Infrared spectrum') -> tuple[plt.Figure, plt.Axes]: # noqa
    '''
    Plots Infrared Spectrum

    Parameters
    ----------
    ir_spec: InfraredSpectrum
        InfraredSpectrum object
    fig: plt.Figure | None, optional
        If provided, uses this Figure object for plotting
    ax: plt.Axes | None, optional
        If provided, plots data to this axis
    label: str, default ''
        Label for plotted line
    color: str, default 'k'
        Colour of plotted line
    line_style: str, default '-'
        Style of plotted line (pyplot ls keyword)
    save: bool, default False
        If True, plot is saved to save_name
    save_name: str | pathlib.Path, default 'ir_spectrum.png'
        If save is True, plot is saved to this location/filename
    show: bool, default False
        If True, plot is shown on screen
    verbose: bool, default True
        If True, plot file location is written to terminal
    window_title: str, default 'Infrared spectrum'
        Title of figure window, not of plot
    legend: bool, default True
        If True, a legend is added to the plot with label(s) given by \n
        ir_spec.comment
    '''

    if fig is None or ax is None:
        width = 4
        width_cm = 10.4
        fig, ax = plt.subplots(
            1,
            1,
            num=window_title,
            figsize=(width, 3.5)
        )
    else:
        width = fig.get_size_inches()[0]
        width_cm = width * 2.54

    # Main spectrum
    ax.plot(
        ir_spec.x_grid,
        ir_spec.y_values,
        label=ir_spec.comment,
        color=color,
        ls=line_style
    )

    ax.set_xlim([np.min(ir_spec.x_grid), np.max(ir_spec.x_grid)])

    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)

    ax.minorticks_on()

    ax.set_xlabel(r'Wavenumber (cm$^\mathregular{-1}$)')
    ax.set_ylabel(ir_spec.y_label_mathmode)

    fig.tight_layout()

    if save:
        plt.savefig(save_name, dpi=500)
        if verbose:
            ut.cprint(
                f'\nSpectrum saved to\n {save_name}',
                'cyan'
            )
            ut.cprint(
                f'Use width={width:.1f} in. or {width_cm:.1f} cm',
                'cyan'
            )

    if show:
        plt.show()

    return fig, ax


def plot_raman(wavenumbers: ArrayLike, intensities: ArrayLike,
               lineshape: str = 'gaussian', linewidth: float = 10.,
               xlim: list[float] = 0., x_unit: str = 'wavenumber',
               abs_type: str = 'absorption', ylim: list[float] = 'auto',
               save: bool = False, save_name: str = 'raman.png',
               show: bool = False,
               window_title: str = 'Raman spectrum'):
    '''
    Plots Raman Spectrum
    '''

    raise NotImplementedError


def plot_cd(save: bool = False, save_name: str = 'raman.png',
            show: bool = False,
            window_title: str = 'Raman spectrum'):
    '''
    Plots circular dichroism data
    '''

    raise NotImplementedError


def plot_ailft_orb_energies(energies: ArrayLike, labels: ArrayLike = None,
                            groups: ArrayLike = None,
                            occupations: ArrayLike = None,
                            y_unit: str = r'cm^{-1}',
                            save: bool = False,
                            save_name: str = 'ai_lft_energies.png',
                            show: bool = False,
                            window_title: str = 'AI-LFT Orbital Energies',
                            verbose: bool = True) -> tuple[plt.Figure, plt.Axes]: # noqa
    '''
    Parameters
    ----------
    energies: array_like
        Energies which are in same unit as y_unit
    labels: array_like | None, optional
        If provided, labels are added next to energy levels.
    groups: array_like | None, optional
        If provided, groups orbitals together by offsetting x coordinate
    occupations: array_like | None, optional
        If provided, each orbital is populated with either 0, 1 or 2 electrons
    y_unit: str, default 'cm^{-1}'
        Mathmode y-unit which matches input chit data
    save: bool, default False
        If True, plot is saved to save_name
    save_name: str, default 'ai_lft_energies.png'
        If save is True, plot is saved to this location/filename
    show: bool, default False
        If True, plot is shown on screen
    window_title: str, default 'AI-LFT Orbital Energies'
        Title of figure window, not of plot
    verbose: bool, default True
        If True, plot file location is written to terminal

    Returns
    -------
    plt.Figure
        Matplotlib Figure object
    plt.Axes
        Matplotlib Axis object for plot
    '''

    width = 3.5
    width_cm = width * 2.54
    fig, ax = plt.subplots(1, 1, figsize=(width, width), num=window_title)
    ax.yaxis.set_minor_locator(AutoMinorLocator())

    if groups is not None:
        groups = list(groups)
        groups = ut.flatten_recursive(groups)
        if len(groups) != len(energies):
            raise ValueError('Number of groups does not match number of states') # noqa
        # Split group by differing value
        groups = [list(x) for _, x in itertools.groupby(groups)]
        # X values for each group
        xvals = [list(range(len(grp))) for grp in groups]
        # Centre each group so that the middle is zero
        xvals = [g - sum(grp)/len(grp) for grp in xvals for g in grp]
    else:
        xvals = [1] * len(energies)

    ax.plot(
        xvals,
        energies,
        lw=0,
        marker='_',
        mew=1.5,
        color='k',
        markersize=25
    )

    if occupations is not None:
        if len(occupations) != len(energies):
            raise ValueError('Number of occupation numbers does not match number of states') # noqa

        # Plot each marker
        _fd = {'size': 20, 'family': 'DejaVu Sans'}
        va = 'center_baseline'
        for occ, en, xval in zip(occupations, energies, xvals):
            lx = xval - 3 / 40
            # Up and down
            if occ == 2:
                ax.text(lx, en, s='↿', color='k', fontdict=_fd, va=va)
                ax.text(lx + 1 / 40, en, s='⇂', color='k', fontdict=_fd, va=va)
            # up
            elif occ == 1:
                ax.text(lx, en, s='↿', color='k', fontdict=_fd, va=va)
            # down
            elif occ == -1:
                ax.text(lx, en, s='⇂', color='k', fontdict=_fd, va=va)

    if labels is not None:
        for xval, energy, label in zip(xvals, energies, labels):
            ax.text(
                xval * 1.05,
                energy,
                rf'${label}$'
            )

    ax.set_xticklabels([])
    ax.set_xticks([])
    _lims = ax.get_xlim()
    if groups is None:
        ax.set_xlim([_lims[0]*0.9, _lims[1]*1.2])
    else:
        ax.set_xlim([_lims[0]*1.2, _lims[1]*1.2])

    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(False)

    ax.set_ylabel(rf'Energy $\mathregular{{({y_unit})}}$')

    fig.tight_layout()

    if save:
        plt.savefig(save_name, dpi=500)
        if verbose:
            ut.cprint(f'\nAI-LFT orbitals saved to\n{save_name}', 'cyan')
            ut.cprint(f'Use width={width:.1f} in. or {width_cm:.1f} cm', 'cyan') # noqa

    if show:
        plt.show()
